# Copyright (C) 2011 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""Resizefs: cloud-config module which resizes the filesystem"""

import errno
import logging
import os
import re
import stat
from textwrap import dedent
from typing import Optional

from cloudinit import subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.distros import ALL_DISTROS
from cloudinit.settings import PER_ALWAYS

NOBLOCK = "noblock"

meta: MetaSchema = {
    "id": "cc_resizefs",
    "name": "Resizefs",
    "title": "Resize filesystem",
    "description": dedent(
        """\
        Resize a filesystem to use all available space on partition. This
        module is useful along with ``cc_growpart`` and will ensure that if the
        root partition has been resized the root filesystem will be resized
        along with it. By default, ``cc_resizefs`` will resize the root
        partition and will block the boot process while the resize command is
        running. Optionally, the resize operation can be performed in the
        background while cloud-init continues running modules. This can be
        enabled by setting ``resize_rootfs`` to ``noblock``. This module can be
        disabled altogether by setting ``resize_rootfs`` to ``false``."""
    ),
    "distros": [ALL_DISTROS],
    "examples": [
        "resize_rootfs: false  # disable root filesystem resize operation",
        "resize_rootfs: noblock  # runs resize operation in the background",
    ],
    "frequency": PER_ALWAYS,
    "activate_by_schema_keys": [],
}

__doc__ = get_meta_doc(meta)

LOG = logging.getLogger(__name__)


def _resize_btrfs(mount_point, devpth):
    # If "/" is ro resize will fail. However it should be allowed since resize
    # makes everything bigger and subvolumes that are not ro will benefit.
    # Use a subvolume that is not ro to trick the resize operation to do the
    # "right" thing. The use of ".snapshot" is specific to "snapper" a generic
    # solution would be walk the subvolumes and find a rw mounted subvolume.
    if not util.mount_is_read_write(mount_point) and os.path.isdir(
        "%s/.snapshots" % mount_point
    ):
        cmd = [
            "btrfs",
            "filesystem",
            "resize",
            "max",
            "%s/.snapshots" % mount_point,
        ]
    else:
        cmd = ["btrfs", "filesystem", "resize", "max", mount_point]

    # btrfs has exclusive operations and resize may fail if btrfs is busy
    # doing one of the operations that prevents resize. As of btrfs 5.10
    # the resize operation can be queued
    btrfs_with_queue = util.Version.from_str("5.10")
    system_btrfs_ver = util.Version.from_str(
        subp.subp(["btrfs", "--version"])[0].split("v")[-1].strip()
    )
    if system_btrfs_ver >= btrfs_with_queue:
        idx = cmd.index("resize")
        cmd.insert(idx + 1, "--enqueue")

    return tuple(cmd)


def _resize_ext(mount_point, devpth):
    return ("resize2fs", devpth)


def _resize_xfs(mount_point, devpth):
    return ("xfs_growfs", mount_point)


def _resize_ufs(mount_point, devpth):
    return ("growfs", "-y", mount_point)


def _resize_zfs(mount_point, devpth):
    return ("zpool", "online", "-e", mount_point, devpth)


def _resize_hammer2(mount_point, devpth):
    return ("hammer2", "growfs", mount_point)


def _resize_bcachefs(mount_point, devpth):
    """Single device resize"""
    return ("bcachefs", "device", "resize", devpth)


def _can_skip_resize_ufs(mount_point, devpth):
    # possible errors cases on the code-path to growfs -N following:
    # https://github.com/freebsd/freebsd/blob/HEAD/sbin/growfs/growfs.c
    # This is the "good" error:
    skip_start = "growfs: requested size"
    skip_contain = "is not larger than the current filesystem size"
    # growfs exits with 1 for almost all cases up to this one.
    # This means we can't just use rcs=[0, 1] as subp parameter:
    try:
        subp.subp(["growfs", "-N", devpth])
    except subp.ProcessExecutionError as e:
        if e.stderr.startswith(skip_start) and skip_contain in e.stderr:
            # This FS is already at the desired size
            return True
        else:
            raise e
    return False


# Do not use a dictionary as these commands should be able to be used
# for multiple filesystem types if possible, e.g. one command for
# ext2, ext3 and ext4.
RESIZE_FS_PREFIXES_CMDS = [
    ("btrfs", _resize_btrfs),
    ("ext", _resize_ext),
    ("xfs", _resize_xfs),
    ("ufs", _resize_ufs),
    ("zfs", _resize_zfs),
    ("hammer2", _resize_hammer2),
    ("bcachefs", _resize_bcachefs),
]

RESIZE_FS_PRECHECK_CMDS = {"ufs": _can_skip_resize_ufs}


def get_device_info_from_zpool(zpool) -> Optional[str]:
    # zpool has 10 second timeout waiting for /dev/zfs LP: #1760173
    log_warn = LOG.debug if util.is_container() else LOG.warning
    if not os.path.exists("/dev/zfs"):
        LOG.debug("Cannot get zpool info, no /dev/zfs")
        return None
    try:
        zpoolstatus, err = subp.subp(["zpool", "status", zpool])
        if err:
            LOG.info(
                "zpool status returned error: [%s] for zpool [%s]",
                err,
                zpool,
            )
            return None
    except subp.ProcessExecutionError as err:
        log_warn("Unable to get zpool status of %s: %s", zpool, err)
        return None
    r = r".*(ONLINE).*"
    for line in zpoolstatus.split("\n"):
        if re.search(r, line) and zpool not in line and "state" not in line:
            disk = line.split()[0]
            LOG.debug('found zpool "%s" on disk %s', zpool, disk)
            return disk
    log_warn(
        "No zpool found: [%s]: out: [%s] err: %s", zpool, zpoolstatus, err
    )
    return None


def can_skip_resize(fs_type, resize_what, devpth):
    fstype_lc = fs_type.lower()
    for i, func in RESIZE_FS_PRECHECK_CMDS.items():
        if fstype_lc.startswith(i):
            return func(resize_what, devpth)
    return False


def maybe_get_writable_device_path(devpath, info):
    """Return updated devpath if the devpath is a writable block device.

    @param devpath: Requested path to the root device we want to resize.
    @param info: String representing information about the requested device.
    @param log: Logger to which logs will be added upon error.

    @returns devpath or updated devpath per kernel commandline if the device
        path is a writable block device, returns None otherwise.
    """
    container = util.is_container()

    # Ensure the path is a block device.
    if (
        devpath == "/dev/root"
        and not os.path.exists(devpath)
        and not container
    ):
        devpath = util.rootdev_from_cmdline(util.get_cmdline())
        if devpath is None:
            LOG.warning("Unable to find device '/dev/root'")
            return None
        LOG.debug("Converted /dev/root to '%s' per kernel cmdline", devpath)

    if devpath == "overlayroot":
        LOG.debug("Not attempting to resize devpath '%s': %s", devpath, info)
        return None

    # FreeBSD zpool can also just use gpt/<label>
    # with that in mind we can not do an os.stat on "gpt/whatever"
    # therefore return the devpath already here.
    if devpath.startswith("gpt/"):
        LOG.debug("We have a gpt label - just go ahead")
        return devpath
    # Alternatively, our device could simply be a name as returned by gpart,
    # such as da0p3
    if not devpath.startswith("/dev/") and not os.path.exists(devpath):
        fulldevpath = "/dev/" + devpath.lstrip("/")
        LOG.debug(
            "'%s' doesn't appear to be a valid device path. Trying '%s'",
            devpath,
            fulldevpath,
        )
        devpath = fulldevpath

    try:
        statret = os.stat(devpath)
    except OSError as exc:
        if container and exc.errno == errno.ENOENT:
            LOG.debug(
                "Device '%s' did not exist in container. cannot resize: %s",
                devpath,
                info,
            )
        elif exc.errno == errno.ENOENT:
            LOG.warning(
                "Device '%s' did not exist. cannot resize: %s", devpath, info
            )
        else:
            raise exc
        return None

    if not stat.S_ISBLK(statret.st_mode) and not stat.S_ISCHR(statret.st_mode):
        if container:
            LOG.debug(
                "device '%s' not a block device in container."
                " cannot resize: %s",
                devpath,
                info,
            )
        else:
            LOG.warning(
                "device '%s' not a block device. cannot resize: %s",
                devpath,
                info,
            )
        return None
    return devpath  # The writable block devpath


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    if len(args) != 0:
        resize_root = args[0]
    else:
        resize_root = util.get_cfg_option_str(cfg, "resize_rootfs", True)
    if not util.translate_bool(resize_root, addons=[NOBLOCK]):
        LOG.debug("Skipping module named %s, resizing disabled", name)
        return

    # TODO(harlowja): allow what is to be resized to be configurable??
    resize_what = "/"
    result = util.get_mount_info(resize_what, LOG)
    if not result:
        LOG.warning("Could not determine filesystem type of %s", resize_what)
        return

    (devpth, fs_type, mount_point) = result

    # if we have a zfs then our device path at this point
    # is the zfs label. For example: vmzroot/ROOT/freebsd
    # we will have to get the zpool name out of this
    # and set the resize_what variable to the zpool
    # so the _resize_zfs function gets the right attribute.
    if fs_type == "zfs":
        zpool = devpth.split("/")[0]
        devpth = get_device_info_from_zpool(zpool)
        if not devpth:
            return  # could not find device from zpool
        resize_what = zpool

    info = "dev=%s mnt_point=%s path=%s" % (devpth, mount_point, resize_what)
    LOG.debug("resize_info: %s", info)

    devpth = maybe_get_writable_device_path(devpth, info)
    if not devpth:
        return  # devpath was not a writable block device

    resizer = None
    if can_skip_resize(fs_type, resize_what, devpth):
        LOG.debug(
            "Skip resize filesystem type %s for %s", fs_type, resize_what
        )
        return

    fstype_lc = fs_type.lower()
    for (pfix, root_cmd) in RESIZE_FS_PREFIXES_CMDS:
        if fstype_lc.startswith(pfix):
            resizer = root_cmd
            break

    if not resizer:
        LOG.warning(
            "Not resizing unknown filesystem type %s for %s",
            fs_type,
            resize_what,
        )
        return

    resize_cmd = resizer(resize_what, devpth)
    LOG.debug(
        "Resizing %s (%s) using %s", resize_what, fs_type, " ".join(resize_cmd)
    )

    if resize_root == NOBLOCK:
        # Fork to a child that will run
        # the resize command
        util.fork_cb(
            util.log_time,
            logfunc=LOG.debug,
            msg="backgrounded Resizing",
            func=do_resize,
            args=(resize_cmd,),
        )
    else:
        util.log_time(
            logfunc=LOG.debug,
            msg="Resizing",
            func=do_resize,
            args=(resize_cmd,),
        )

    action = "Resized"
    if resize_root == NOBLOCK:
        action = "Resizing (via forking)"
    LOG.debug(
        "%s root filesystem (type=%s, val=%s)", action, fs_type, resize_root
    )


def do_resize(resize_cmd):
    try:
        subp.subp(resize_cmd)
    except subp.ProcessExecutionError:
        util.logexc(LOG, "Failed to resize filesystem (cmd=%s)", resize_cmd)
        raise
    # TODO(harlowja): Should we add a fsck check after this to make
    # sure we didn't corrupt anything?



from functools import partial
from itertools import count
from unittest import mock

import pytest

from cloudinit.config.cc_phone_home import POST_LIST_ALL, handle
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import skipUnlessJsonSchema
from tests.unittests.util import get_cloud

phone_home = partial(handle, name="test", cloud=get_cloud(), args=[])


@pytest.fixture(autouse=True)
def common_mocks(mocker):
    mocker.patch("cloudinit.util.load_text_file", side_effect=count())


@mock.patch("cloudinit.url_helper.readurl")
class TestPhoneHome:
    def test_default_call(self, m_readurl):
        cfg = {"phone_home": {"url": "myurl"}}
        phone_home(cfg=cfg)
        assert m_readurl.call_args == mock.call(
            "myurl",
            data={
                "pub_key_rsa": "0",
                "pub_key_ecdsa": "1",
                "pub_key_ed25519": "2",
                "instance_id": "iid-datasource-none",
                "hostname": "hostname",
                "fqdn": "hostname",
            },
            retries=9,
            sec_between=3,
            ssl_details={},
        )

    def test_no_url(self, m_readurl, caplog):
        cfg = {"phone_home": {}}
        phone_home(cfg=cfg)
        assert "Skipping module named" in caplog.text
        assert m_readurl.call_count == 0

    @pytest.mark.parametrize(
        "tries, expected_retries",
        [
            (-1, -2),
            (0, -1),
            (1, 0),
            (2, 1),
            ("2", 1),
            ("two", 9),
            (None, 9),
            ({}, 9),
        ],
    )
    def test_tries(self, m_readurl, tries, expected_retries, caplog):
        cfg = {"phone_home": {"url": "dontcare"}}
        if tries is not None:
            cfg["phone_home"]["tries"] = tries
        phone_home(cfg=cfg)
        assert m_readurl.call_args[1]["retries"] == expected_retries

    def test_post_all(self, m_readurl):
        cfg = {"phone_home": {"url": "test", "post": "all"}}
        phone_home(cfg=cfg)
        for key in POST_LIST_ALL:
            assert key in m_readurl.call_args[1]["data"]

    def test_custom_post_list(self, m_readurl):
        post_list = ["pub_key_rsa, hostname"]
        cfg = {"phone_home": {"url": "test", "post": post_list}}
        phone_home(cfg=cfg)
        for key in post_list:
            assert key in m_readurl.call_args[1]["data"]
        assert len(m_readurl.call_args[1]["data"]) == len(post_list)

    def test_invalid_post(self, m_readurl, caplog):
        post_list = ["spam", "hostname"]
        cfg = {"phone_home": {"url": "test", "post": post_list}}
        phone_home(cfg=cfg)
        assert "hostname" in m_readurl.call_args[1]["data"]
        assert m_readurl.call_args[1]["data"]["spam"] == "N/A"
        assert (
            "spam from 'post' configuration list not available" in caplog.text
        )


class TestPhoneHomeSchema:
    @pytest.mark.parametrize(
        "config",
        [
            # phone_home definition with url
            {"phone_home": {"post": ["pub_key_rsa"]}},
            # post using string other than "all"
            {"phone_home": {"url": "test_url", "post": "pub_key_rsa"}},
            # post using list with misspelled entry
            {"phone_home": {"url": "test_url", "post": ["pub_kye_rsa"]}},
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config):
        with pytest.raises(SchemaValidationError):
            validate_cloudconfig_schema(config, get_schema(), strict=True)



# This file is part of cloud-init. See LICENSE file for license information.
import re
from copy import deepcopy
from unittest import mock

import pytest

from cloudinit.config import cc_lxd
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from cloudinit.helpers import Paths
from cloudinit.util import del_file
from tests.unittests import helpers as t_help
from tests.unittests.util import get_cloud

BACKEND_DEF = (
    ("zfs", "zfs", "zfsutils-linux"),
    ("btrfs", "mkfs.btrfs", "btrfs-progs"),
    ("lvm", "lvcreate", "lvm2"),
    ("dir", None, None),
)
LXD_INIT_CFG = {
    "lxd": {
        "init": {
            "network_address": "0.0.0.0",
            "storage_backend": "zfs",
            "storage_pool": "poolname",
        }
    }
}


class TestLxd(t_help.CiTestCase):

    with_logs = True

    @mock.patch("cloudinit.config.cc_lxd.util.system_info")
    @mock.patch("cloudinit.config.cc_lxd.os.path.exists", return_value=True)
    @mock.patch("cloudinit.config.cc_lxd.subp.subp", return_value=True)
    @mock.patch("cloudinit.config.cc_lxd.subp.which", return_value=False)
    @mock.patch(
        "cloudinit.config.cc_lxd.maybe_cleanup_default", return_value=None
    )
    def test_lxd_init(self, maybe_clean, which, subp, exists, system_info):
        system_info.return_value = {"uname": [0, 1, "mykernel"]}
        tmpdir = self.tmp_dir()
        sem_file = f"{tmpdir}/sem/snap_seeded.once"
        cc = get_cloud(mocked_distro=True, paths=Paths({"cloud_dir": tmpdir}))
        install = cc.distro.install_packages

        for backend, cmd, package in BACKEND_DEF:
            lxd_cfg = deepcopy(LXD_INIT_CFG)
            lxd_cfg["lxd"]["init"]["storage_backend"] = backend
            subp.call_args_list = []
            install.call_args_list = []
            exists.call_args_list = []
            cc_lxd.handle("cc_lxd", lxd_cfg, cc, [])
            if cmd:
                which.assert_called_with(cmd)
            # no bridge config, so maybe_cleanup should not be called.
            self.assertFalse(maybe_clean.called)
            if package:
                self.assertEqual(
                    [mock.call([package])],
                    install.call_args_list,
                )
            self.assertEqual(
                [
                    mock.call(["snap", "install", "lxd"]),
                    mock.call(["lxd", "waitready", "--timeout=300"]),
                    mock.call(
                        [
                            "lxd",
                            "init",
                            "--auto",
                            "--network-address=0.0.0.0",
                            f"--storage-backend={backend}",
                            "--storage-pool=poolname",
                        ]
                    ),
                ],
                subp.call_args_list,
            )

            if backend == "lvm":
                self.assertEqual(
                    [
                        mock.call(sem_file),
                        mock.call(
                            "/lib/modules/mykernel/"
                            "kernel/drivers/md/dm-thin-pool.ko"
                        ),
                    ],
                    exists.call_args_list,
                )
            else:
                self.assertEqual([mock.call(sem_file)], exists.call_args_list)
            del_file(sem_file)

    @mock.patch("cloudinit.config.cc_lxd.maybe_cleanup_default")
    @mock.patch("cloudinit.config.cc_lxd.subp")
    @mock.patch("cloudinit.config.cc_lxd.subp.which", return_value=False)
    def test_lxd_install(self, m_which, mock_subp, m_maybe_clean):
        cc = get_cloud(paths=Paths({"cloud_dir": self.tmp_dir()}))
        cc.distro = mock.MagicMock()
        mock_subp.which.return_value = None
        cc_lxd.handle("cc_lxd", LXD_INIT_CFG, cc, [])
        self.assertNotIn("WARN", self.logs.getvalue())
        self.assertTrue(cc.distro.install_packages.called)
        cc_lxd.handle("cc_lxd", LXD_INIT_CFG, cc, [])
        self.assertFalse(m_maybe_clean.called)
        install_pkg = cc.distro.install_packages.call_args_list[0][0][0]
        self.assertEqual(sorted(install_pkg), ["zfsutils-linux"])

    @mock.patch("cloudinit.config.cc_lxd.maybe_cleanup_default")
    @mock.patch("cloudinit.config.cc_lxd.subp")
    def test_no_init_does_nothing(self, mock_subp, m_maybe_clean):
        cc = get_cloud(paths=Paths({"cloud_dir": self.tmp_dir()}))
        cc.distro = mock.MagicMock()
        cc_lxd.handle("cc_lxd", {"lxd": {}}, cc, [])
        self.assertFalse(cc.distro.install_packages.called)
        self.assertFalse(mock_subp.subp.called)
        self.assertFalse(m_maybe_clean.called)

    @mock.patch("cloudinit.config.cc_lxd.maybe_cleanup_default")
    @mock.patch("cloudinit.config.cc_lxd.subp")
    def test_no_lxd_does_nothing(self, mock_subp, m_maybe_clean):
        cc = get_cloud(paths=Paths({"cloud_dir": self.tmp_dir()}))
        cc.distro = mock.MagicMock()
        cc_lxd.handle("cc_lxd", {"package_update": True}, cc, [])
        self.assertFalse(cc.distro.install_packages.called)
        self.assertFalse(mock_subp.subp.called)
        self.assertFalse(m_maybe_clean.called)

    @mock.patch("cloudinit.config.cc_lxd.util.wait_for_snap_seeded")
    @mock.patch("cloudinit.config.cc_lxd.subp")
    def test_lxd_preseed(self, mock_subp, wait_for_snap_seeded):
        cc = get_cloud(paths=Paths({"cloud_dir": self.tmp_dir()}))
        cc.distro = mock.MagicMock()
        cc_lxd.handle(
            "cc_lxd",
            {"lxd": {"preseed": '{"chad": True}'}},
            cc,
            [],
        )
        self.assertEqual(
            [
                mock.call(["lxd", "waitready", "--timeout=300"]),
                mock.call(["lxd", "init", "--preseed"], data='{"chad": True}'),
            ],
            mock_subp.subp.call_args_list,
        )
        wait_for_snap_seeded.assert_called_once_with(cc)

    def test_lxd_debconf_new_full(self):
        data = {
            "mode": "new",
            "name": "testbr0",
            "ipv4_address": "10.0.8.1",
            "ipv4_netmask": "24",
            "ipv4_dhcp_first": "10.0.8.2",
            "ipv4_dhcp_last": "10.0.8.254",
            "ipv4_dhcp_leases": "250",
            "ipv4_nat": "true",
            "ipv6_address": "fd98:9e0:3744::1",
            "ipv6_netmask": "64",
            "ipv6_nat": "true",
            "domain": "lxd",
        }
        self.assertEqual(
            cc_lxd.bridge_to_debconf(data),
            {
                "lxd/setup-bridge": "true",
                "lxd/bridge-name": "testbr0",
                "lxd/bridge-ipv4": "true",
                "lxd/bridge-ipv4-address": "10.0.8.1",
                "lxd/bridge-ipv4-netmask": "24",
                "lxd/bridge-ipv4-dhcp-first": "10.0.8.2",
                "lxd/bridge-ipv4-dhcp-last": "10.0.8.254",
                "lxd/bridge-ipv4-dhcp-leases": "250",
                "lxd/bridge-ipv4-nat": "true",
                "lxd/bridge-ipv6": "true",
                "lxd/bridge-ipv6-address": "fd98:9e0:3744::1",
                "lxd/bridge-ipv6-netmask": "64",
                "lxd/bridge-ipv6-nat": "true",
                "lxd/bridge-domain": "lxd",
            },
        )

    def test_lxd_debconf_new_partial(self):
        data = {
            "mode": "new",
            "ipv6_address": "fd98:9e0:3744::1",
            "ipv6_netmask": "64",
            "ipv6_nat": "true",
        }
        self.assertEqual(
            cc_lxd.bridge_to_debconf(data),
            {
                "lxd/setup-bridge": "true",
                "lxd/bridge-ipv6": "true",
                "lxd/bridge-ipv6-address": "fd98:9e0:3744::1",
                "lxd/bridge-ipv6-netmask": "64",
                "lxd/bridge-ipv6-nat": "true",
            },
        )

    def test_lxd_debconf_existing(self):
        data = {"mode": "existing", "name": "testbr0"}
        self.assertEqual(
            cc_lxd.bridge_to_debconf(data),
            {
                "lxd/setup-bridge": "false",
                "lxd/use-existing-bridge": "true",
                "lxd/bridge-name": "testbr0",
            },
        )

    def test_lxd_debconf_none(self):
        data = {"mode": "none"}
        self.assertEqual(
            cc_lxd.bridge_to_debconf(data),
            {"lxd/setup-bridge": "false", "lxd/bridge-name": ""},
        )

    def test_lxd_cmd_new_full(self):
        data = {
            "mode": "new",
            "name": "testbr0",
            "ipv4_address": "10.0.8.1",
            "ipv4_netmask": "24",
            "ipv4_dhcp_first": "10.0.8.2",
            "ipv4_dhcp_last": "10.0.8.254",
            "ipv4_dhcp_leases": "250",
            "ipv4_nat": "true",
            "ipv6_address": "fd98:9e0:3744::1",
            "ipv6_netmask": "64",
            "ipv6_nat": "true",
            "domain": "lxd",
            "mtu": 9000,
        }
        self.assertEqual(
            cc_lxd.bridge_to_cmd(data),
            (
                [
                    "network",
                    "create",
                    "testbr0",
                    "ipv4.address=10.0.8.1/24",
                    "ipv4.nat=true",
                    "ipv4.dhcp.ranges=10.0.8.2-10.0.8.254",
                    "ipv6.address=fd98:9e0:3744::1/64",
                    "ipv6.nat=true",
                    "dns.domain=lxd",
                    "bridge.mtu=9000",
                ],
                ["network", "attach-profile", "testbr0", "default", "eth0"],
            ),
        )

    def test_lxd_cmd_new_partial(self):
        data = {
            "mode": "new",
            "ipv6_address": "fd98:9e0:3744::1",
            "ipv6_netmask": "64",
            "ipv6_nat": "true",
            "mtu": -1,
        }
        self.assertEqual(
            cc_lxd.bridge_to_cmd(data),
            (
                [
                    "network",
                    "create",
                    "lxdbr0",
                    "ipv4.address=none",
                    "ipv6.address=fd98:9e0:3744::1/64",
                    "ipv6.nat=true",
                ],
                ["network", "attach-profile", "lxdbr0", "default", "eth0"],
            ),
        )

    def test_lxd_cmd_existing(self):
        data = {"mode": "existing", "name": "testbr0"}
        self.assertEqual(
            cc_lxd.bridge_to_cmd(data),
            (
                None,
                ["network", "attach-profile", "testbr0", "default", "eth0"],
            ),
        )

    def test_lxd_cmd_none(self):
        data = {"mode": "none"}
        self.assertEqual(cc_lxd.bridge_to_cmd(data), (None, None))


class TestLxdMaybeCleanupDefault(t_help.CiTestCase):
    """Test the implementation of maybe_cleanup_default."""

    defnet = cc_lxd._DEFAULT_NETWORK_NAME

    @mock.patch("cloudinit.config.cc_lxd._lxc")
    def test_network_other_than_default_not_deleted(self, m_lxc):
        """deletion or removal should only occur if bridge is default."""
        cc_lxd.maybe_cleanup_default(
            net_name="lxdbr1", did_init=True, create=True, attach=True
        )
        m_lxc.assert_not_called()

    @mock.patch("cloudinit.config.cc_lxd._lxc")
    def test_did_init_false_does_not_delete(self, m_lxc):
        """deletion or removal should only occur if did_init is True."""
        cc_lxd.maybe_cleanup_default(
            net_name=self.defnet, did_init=False, create=True, attach=True
        )
        m_lxc.assert_not_called()

    @mock.patch("cloudinit.config.cc_lxd._lxc")
    def test_network_deleted_if_create_true(self, m_lxc):
        """deletion of network should occur if create is True."""
        cc_lxd.maybe_cleanup_default(
            net_name=self.defnet, did_init=True, create=True, attach=False
        )
        m_lxc.assert_called_with(["network", "delete", self.defnet])

    @mock.patch("cloudinit.config.cc_lxd._lxc")
    def test_device_removed_if_attach_true(self, m_lxc):
        """deletion of network should occur if create is True."""
        nic_name = "my_nic"
        profile = "my_profile"
        cc_lxd.maybe_cleanup_default(
            net_name=self.defnet,
            did_init=True,
            create=False,
            attach=True,
            profile=profile,
            nic_name=nic_name,
        )
        m_lxc.assert_called_once_with(
            ["profile", "device", "remove", profile, nic_name]
        )


class TestGetRequiredPackages:
    @pytest.mark.parametrize(
        "storage_type, cmd, preseed, package",
        (
            ("zfs", "zfs", "", "zfsutils-linux"),
            ("btrfs", "mkfs.btrfs", "", "btrfs-progs"),
            ("lvm", "lvcreate", "", "lvm2"),
            ("lvm", "lvcreate", "storage_pools: [{driver: lvm}]", "lvm2"),
            ("dir", None, "", None),
        ),
    )
    @mock.patch("cloudinit.config.cc_lxd.subp.which", return_value=False)
    def test_lxd_package_install(
        self, m_which, storage_type, cmd, preseed, package
    ):
        if preseed:  # preseed & lxd.init mutually exclusive
            init_cfg = {}
        else:
            lxd_cfg = deepcopy(LXD_INIT_CFG)
            lxd_cfg["lxd"]["init"]["storage_backend"] = storage_type
            init_cfg = lxd_cfg["lxd"]["init"]

        packages = cc_lxd.get_required_packages(init_cfg, preseed)
        which_calls = []
        if package:
            which_calls.append(mock.call(cmd))
            assert package in packages
        assert which_calls == m_which.call_args_list


class TestLXDSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Only allow init, bridge and preseed keys
            ({"lxd": {"bridgeo": 1}}, "Additional properties are not allowed"),
            # Only allow init.storage_backend values zfs and dir
            (
                {"lxd": {"init": {"storage_backend": "1zfs"}}},
                re.escape("not one of ['zfs', 'dir', 'lvm', 'btrfs']"),
            ),
            ({"lxd": {"init": {"storage_backend": "lvm"}}}, None),
            ({"lxd": {"init": {"storage_backend": "btrfs"}}}, None),
            ({"lxd": {"init": {"storage_backend": "zfs"}}}, None),
            # Require bridge.mode
            ({"lxd": {"bridge": {}}}, "bridge: 'mode' is a required property"),
            # Require init or bridge keys
            ({"lxd": {}}, f"lxd: {{}} {t_help.SCHEMA_EMPTY_ERROR}"),
            # Require some non-empty preseed config of type string
            ({"lxd": {"preseed": {}}}, "not of type 'string'"),
            ({"lxd": {"preseed": ""}}, None),
            ({"lxd": {"preseed": "this is {} opaque"}}, None),
            # Require bridge.mode
            ({"lxd": {"bridge": {"mode": "new", "mtu": 9000}}}, None),
            # LXD's default value
            ({"lxd": {"bridge": {"mode": "new", "mtu": -1}}}, None),
            # No additionalProperties
            (
                {"lxd": {"init": {"invalid": None}}},
                "Additional properties are not allowed",
            ),
            (
                {"lxd": {"bridge": {"mode": None, "garbage": None}}},
                "Additional properties are not allowed",
            ),
        ],
    )
    @t_help.skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            validate_cloudconfig_schema(config, get_schema(), strict=True)

    @pytest.mark.parametrize(
        "init_cfg, bridge_cfg, preseed_str, error_expectation",
        (
            pytest.param(
                {}, {}, "", t_help.does_not_raise(), id="empty_cfgs_no_errors"
            ),
            pytest.param(
                {"init-cfg": 1},
                {"bridge-cfg": 2},
                "",
                t_help.does_not_raise(),
                id="cfg_init_and_bridge_allowed",
            ),
            pytest.param(
                {},
                {},
                "profiles: []",
                t_help.does_not_raise(),
                id="cfg_preseed_allowed_without_bridge_or_init",
            ),
            pytest.param(
                {"init-cfg": 1},
                {"bridge-cfg": 2},
                "profiles: []",
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "Unable to configure LXD. lxd.preseed config can not"
                        " be provided with key(s): lxd.init, lxd.bridge"
                    ),
                ),
            ),
            pytest.param(
                "nope",
                {},
                "",
                pytest.raises(
                    ValueError,
                    match=re.escape(
                        "lxd.init config must be a dictionary. found a 'str'"
                    ),
                ),
            ),
        ),
    )
    def test_supplemental_schema_validation_raises_value_error(
        self, init_cfg, bridge_cfg, preseed_str, error_expectation
    ):
        """LXD is strict on invalid user-data raising conspicuous ValueErrors
        cc_lxd.supplemental_schema_validation

        Hard errors result is faster triage/awareness of config problems than
        warnings do.
        """
        with error_expectation:
            cc_lxd.supplemental_schema_validation(
                init_cfg, bridge_cfg, preseed_str
            )



"""Integration test for gh-668.

Ensure that static route to host is working correctly.
The original problem is specific to the ENI renderer but that test is suitable
for all network configuration outputs.
"""

import pytest

from tests.integration_tests import random_mac_address
from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.integration_settings import PLATFORM

DESTINATION_IP = "172.16.0.10"
GATEWAY_IP = "10.0.0.100"
MAC_ADDRESS = random_mac_address()

NETWORK_CONFIG = """\
version: 2
ethernets:
  eth0:
    addresses: [10.0.0.10/8]
    dhcp4: false
    routes:
    - to: {}/32
      via: {}
""".format(
    DESTINATION_IP, GATEWAY_IP
)

EXPECTED_ROUTE = "{} via {}".format(DESTINATION_IP, GATEWAY_IP)


NETWORK_CONFIG_VM = f"""\
{NETWORK_CONFIG}
    match:
      macaddress: {MAC_ADDRESS}
"""

# On netplan 0.106 and above, netplan emits MAC address matches as
# PermanentMACAdddress= on networkd systems. This match clause will
# not match LXD's veth devices which results in no routes being brought up.
# As a result, we match container by NIC name. Leave lxd_vm as MAC match.
# LP: #2022947 represents this netplan behavior
NETWORK_CONFIG_CONTAINER = f"""\
{NETWORK_CONFIG}
    match:
      name: eth0
"""


@pytest.mark.skipif(
    PLATFORM not in ["lxd_container", "lxd_vm"],
    reason="Test requires custom networking provided by LXD",
)
def test_static_route_to_host(session_cloud: IntegrationCloud):
    if PLATFORM == "lxd_vm":
        config_dict = {
            "user.network-config": NETWORK_CONFIG_VM,
            "volatile.eth0.hwaddr": MAC_ADDRESS,
        }
    else:
        config_dict = {"user.network-config": NETWORK_CONFIG_CONTAINER}
    with session_cloud.launch(
        launch_kwargs={
            "execute_via_ssh": False,
            "config_dict": config_dict,
        }
    ) as client:
        route = client.execute("ip route | grep {}".format(DESTINATION_IP))
        assert route.startswith(EXPECTED_ROUTE)



"""Integration test for LP: #1901011

Ensure an ephemeral disk exists after boot.

See https://github.com/canonical/cloud-init/pull/800
"""
import pytest

from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.integration_settings import PLATFORM


@pytest.mark.skipif(PLATFORM != "azure", reason="Test is Azure specific")
@pytest.mark.parametrize(
    "instance_type,is_ephemeral",
    [
        ("Standard_DS1_v2", True),
        ("Standard_D2s_v4", False),
    ],
)
def test_ephemeral(
    instance_type, is_ephemeral, session_cloud: IntegrationCloud, setup_image
):
    if is_ephemeral:
        expected_log = (
            "Ephemeral resource disk '/dev/disk/cloud/azure_resource' exists. "
            "Merging default Azure cloud ephemeral disk configs."
        )
    else:
        expected_log = (
            "Ephemeral resource disk '/dev/disk/cloud/azure_resource' does "
            "not exist. Not merging default Azure cloud ephemeral disk "
            "configs."
        )

    with session_cloud.launch(
        launch_kwargs={"instance_type": instance_type}
    ) as client:
        # Verify log file
        log = client.read_from_file("/var/log/cloud-init.log")
        assert expected_log in log

        # Verify devices
        dev_links = client.execute("ls /dev/disk/cloud")
        assert "azure_root" in dev_links
        assert "azure_root-part1" in dev_links
        if is_ephemeral:
            assert "azure_resource" in dev_links
            assert "azure_resource-part1" in dev_links

        # Verify mounts
        blks = client.execute("lsblk -pPo NAME,TYPE,MOUNTPOINT")
        root_device = client.execute(
            "realpath /dev/disk/cloud/azure_root-part1"
        )
        assert (
            'NAME="{}" TYPE="part" MOUNTPOINT="/"'.format(root_device) in blks
        )
        if is_ephemeral:
            ephemeral_device = client.execute(
                "realpath /dev/disk/cloud/azure_resource-part1"
            )
            assert (
                'NAME="{}" TYPE="part" MOUNTPOINT="/mnt"'.format(
                    ephemeral_device
                )
                in blks
            )
